/*
	
					The Hacker's Diet

					Monthly Log Form
					 
		This file contains all the code for displaying and
		interacting with the Month Log form, which shows the
		detailed data for a given month.	
*/

#include "Hackdiet.h"

/*	With the advent of the PalmOS 3.3 debug ROM, code which functioned
	perfectly for millions of gremlin events started blowing loose due
	to "field validation" reporting discrepancies in text length,
	insertion point, null pointers, etc.  There appear to be two
	principal causes of this.  First of all, even though you're editing
	a field in a table, some PalmOS code, Find for example, looks at a
	static FieldType structure embedded in the active table.  To work
	around this zany behaviour, it's necessary to copy your field into
	the field in the table before making any system call which might
	take it upon itself to validate the static field in the table or
	use it in any way.  The set of such system calls can be determined
	only empirically, and my defensive code is probably more conservative
	than it needs to be the cost in performance is nil.
	
	Second, unless I'm really missing something obvious, it appears that
	on certain occasions the field code in PalmOS manages to get the
	text length and insertion point in a field out of sync with the
	actual contents of the field, which requires defensive code to
	re-set these items if they point beyond the end of the actual
	string, and the defensive code itself requires more defensive code
	to guard against the possibility of a null text handle in the
	field, which can also happen under certain screwball circumstances.
	
	None of these problems are evident except in gremlin testing of
	50,000 events or more, so whatever the underlying cause may be,
	the circumstances which trigger it are extremely rare.  All of
	the defensive code is conditional on GNARLY, allowing it to be
	identified and, hopefully at some point in the future, removed.
	A build without GNARLY will run without errors on the Palm IIIx
	debug ROM.  */

#define GNARLY

//	Application globals

UInt16 logGotoMonth = 0;				// Which month to show (database record index)
UInt16 showSpecificDay = 0;				// Initally make this day (1-31) visible if nonzero
Boolean goToToday = false;				// Go to current day and edit weight ?
Boolean showToday = false;				// Make current day visible
rung recordRung = 0;					// Record rung for today
static char formTitle[MAX_MONTH_YYYY];	// Form title (static because not copied)
static struct hdMonth *m = NULL;		// Unpacked database for month
static UInt16 tableRows = 0;				// Rows in table
static Boolean dirty = false;			// Has content of table been modified ?
static Boolean editing = false;			// True if editing in field
static FieldPtr currentFld = NULL;		// Current field we're editing
static int editRow = -1, editCol = -1;	// If >= 0, current row and column we're editing
static int logCurrentMonth;				// Actual month we're editing
static Boolean logUnlocked;				// True unless private logs hidden by Security application
static int tempUnlock = -1;				// Row of temporarily unlocked cell
static int tempUnlockFirst = -1;		// Flag first tap of temporary unlock
static Boolean tablePaint = true;		// OK to paint fields in table ?
static Boolean tablePainted = false;	// Has table been painted at initial form draw time ?

//	Form globals

static int topDay = 1;					// Top day in scrolling area

#define nColumns	6					// Columns in monthly log table:
#define cDate		0					//		Date
#define cWeight		1					//		Weight
#define cVar		2					//		Variance
#define cRung		3					//		Rung
#define cFlag		4					//		Flag
#define cComment	5					//		Comment

#define RowDay ((topDay - 1) + row)	// Index (0-30) of day for current row
/*	Is column editable?  This applies only to columns edited as text field, not
	the check box cFlag column.  */
#define columnEditable(col) (((col) == cWeight) || ((col) == cRung) || ((col) == cComment))

/*	The rationale for the "dirty field" mechanism is to allow users to
	correct data entry errors when protection of log entries is in
	effect.  When protection (prefs.protectEntries) is enabled, we
	don't allow the user to change values in the future and non-blank
	values on days in the past.  The intent is to protect against an
	inadvertent tap opening a field for editing which might accidentally
	alter it.  But many users keep logs on paper and transfer them to
	the handheld several days at a time.  Since values are being entered
	into blank fields, protection doesn't interfere with this, but
	if the user discovers an error after closing a field, there's no
	way to correct it without turning off protection since it's now a
	non-blank field in the past, and hence protected.  To avoid this
	annoyance, we keep track of fields which were changed since this
	month's log was displayed.  Any field which has been entered in
	this editing session is not considered protected and can be
	modified subsequently in the session.  Once the log is closed,
	however, the field becomes protected.  */

static unsigned char fldDirty[31];	//	Individual field dirty flags

#define MarkFieldDirty(day, column)	fldDirty[day] |= (1 << (column))
#define IsFieldDirty(day, column)	((fldDirty[day] & (1 << (column))) != 0)

//	Forward functions

static void drawDate(MemPtr table, Int16 row, Int16 col, RectanglePtr rect);

//	clearUnlock  --  Clear temporary unlock of row in table

static void clearUnlock(void)
{
	if (tempUnlock >= 0) {
		RectangleType r;
		FormPtr f = FrmGetActiveForm();
		TablePtr tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
		int row = tempUnlock;
		
		tempUnlock = -1;				// Clear temporary unlock
		TblGetItemBounds(tp, row, cDate, &r);
		drawDate(tp, row, cDate, &r);	// Repaint cell in normal mode
	}
	tempUnlockFirst = -1;
}

//	setScrollbar  --  Set scrollbar parameters for active view

static void setScrollbar(void)
{
	FormPtr frm = FrmGetActiveForm();
	ScrollBarPtr scroll = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MonthlyLogDayScrollBar));
	
	if (m != NULL) {
		SclSetScrollBar(scroll, topDay - 1, 0, m->days - tableRows, tableRows - 1);
	}
}

//	scrollTable  --  Repaint table to reflect change to scrollbar

static void scrollTable(void)
{
	int i;
	FormPtr f = FrmGetActiveForm();
	TablePtr tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));

	clearUnlock();
	TblUnhighlightSelection(tp);

	//	Set the checkbox items in the Flags column to current scroll position
	
	for (i = 0; i < tableRows; i++) {
		UInt16 d = i + (topDay - 1);
					
		TblSetItemInt(tp, i, cFlag, d < m->days ? m->flagsArray[d] : 0);
	}
	TblEraseTable(tp);
	TblMarkTableInvalid(tp);
	TblDrawTable(tp);
}

//	editWeight  --  Edit a fixed decimal weight to a decimal value

void editWeight(char *txt, weight w)
{
	int whole, frac;
	
	if (prefs.displayUnit == wStone) {
		/*	If we're editing the weight as stones, display it as
			a whole number of stones (14 pounds) and pounds, separated
			by a space.  For the benefit of Dietcalc.c, if the quantity
			of full stones is zero, the pound quantity is displayed with
			one decimal place unless the decimal would be zero.  */
		whole = (w + 50) / 100;
		if ((whole / 14) == 0) {
			w += 5;
			whole = w / 100;
			frac = (w / 10) % 10;
			if (frac != 0) {
				StrPrintF(txt, "%d %d.%d", whole / 14, whole, frac);
			} else {
				StrPrintF(txt, "%d %d", whole / 14, whole % 14);
			}			
		} else {
			StrPrintF(txt, "%d %d", whole / 14, whole % 14);
		}
	} else {
		w += 5;
		whole = w / 100;
		frac = (w / 10) % 10;
		StrPrintF(txt, "%d.%d", whole, frac);
	}
	NumOut(txt);
}

//	formatField  --  Edit the content of a field into a string

static void formatField(char *txt, UInt16 row, UInt16 col, UInt16 *max)
{
	txt[0] = 0;					// Blank string is the default
	
	switch (col) {
		case cWeight:			// Weight
			//	If it's stones, have to make room for "14 12.7", for example
			*max = (prefs.displayUnit == wStone) ? 7 : 5;
			if ((RowDay < m->days) && (m->weightArray[RowDay] > 0)) {
				editWeight(txt, cvtWeight(m->weightArray[RowDay], (weightunit) m->weightUnit, prefs.displayUnit));
			}
			break;
			
		case cRung:				// Rung
			*max = 2;
			if ((RowDay < m->days) && (m->rungArray[RowDay] != RUNG_ABSENT)) {
				StrPrintF(txt, "%d", m->rungArray[RowDay]);
			}
			break;
			
		case cComment:			// Comment
			*max = MAX_COMMENT;
			if ((RowDay < m->days) && (m->commentArray[RowDay] != 0)) {
				StrCopy(txt, ((char *) m) + m->commentArray[RowDay]);
			}
			break;
			
#ifndef NDEBUG
		default:
			assert(false);
			break;
#endif
	}
}

/*	updateFromField  --  Update cell in the in-memory log from the contents of a field.
						 The return value is true if change to the weight requires
						 repainting the entire trend/variance column.  */

static Boolean updateFromField(char *txt, UInt16 row, UInt16 col)
{
	char copiedField[MAX_COMMENT + 2];
	Boolean trendChange = false, dotD = false,
			DdotD = false, stoneDDdotD = false;
	UInt16 txtlen = StrLen(txt);
	char *decpos;
	
	/*	The following code implements the gimmick which permits copying of
		the last non-blank weight, rung, or comment in the monthly log simply
		by writing a single period in the field (which is easy to do, since it's
		just a double tap in the Graffiti area).  If for some reason you must
		enter a comment consisting of a single period, simply write a space
		after it; that will disable the copying.  */
	
	if ((StrCompare(txt, ".") == 0) ||									//	"."

		(dotD = ((col == cWeight) && (txt[0] == numberDecimal)  &&		//	".d"
		 (txtlen == 2) && isDigit(txt[1]))) ||

		(DdotD = ((col == cWeight) && (txt[1] == numberDecimal)  &&		//	"d.d" or "d."
		 isDigit(txt[0]) &&
		 (((txtlen == 3) && isDigit(txt[2])) ||
		  (txtlen == 2)))) ||

		(stoneDDdotD = ((col == cWeight) &&								//	Stones: "dd.d" or "dd."
		 (prefs.displayUnit == wStone) &&
		 (txt[2] == numberDecimal)  &&
		 isDigit(txt[0]) && isDigit(txt[1]) &&
		 (((txtlen == 4) && isDigit(txt[3])) ||
		  (txtlen == 3))))
	   ) {
		int i;
		
		switch (col) {
			case cWeight:
				for (i = RowDay - 1; i >= 0; i--) {
					if (m->weightArray[i] != 0) {
						editWeight(copiedField, cvtWeight(m->weightArray[i], (weightunit) m->weightUnit, prefs.displayUnit));
						if (dotD) {
						
							/*	User has written ".d" where "." is the decimal character
								from the preferences and d is a digit.  We replace the
								decimal place in the copied field with the digit d.  */
							
							decpos = StrChr(copiedField, numberDecimal);
							if (decpos == NULL) {
								/*	If there's no decimal in the field we're copying
									append the given decimal.  This only happens when
									entering decimal pounds in stone and pound display
									units.  */
								decpos = copiedField + StrLen(copiedField);
							}
							StrCopy(decpos, txt);
						} else if (DdotD) {
						
							/*	User has written "u.d" where "." is the decimal character
								from the preferences and u and d are digits.  We replace the
								units and decimal places of the number with "u.d".  */
							
							decpos = StrChr(copiedField, numberDecimal);
							if (decpos == NULL) {
								/*	If there's no decimal in the field we're copying
									replace the last digit and append the decimal.  This
									only happens when the display unit is stones.  */
								decpos = copiedField + StrLen(copiedField);
							}
							if (decpos > copiedField) {
								decpos--;
								
								/*	One more little twist.  If the display units are stones
									and the previous entry contained a two digit pounds
									field and the units digit in this entry is four or greater,
									we assume the user wishes to replace the pounds field with
									the u.d value entered rather than enter an invalid stones
									and pounds value.  This allows a user who oscillates back
									and forth around x stones 10 pounds to replace the double
									digit value with the single just by writing the the number
									of pounds followed by a period.  */
									
								if ((prefs.displayUnit == wStone) &&
									(decpos > copiedField) &&
									isDigit(*(decpos - 1)) &&
									(txt[0] >= '4') && (txt[0] <= '9')) {
									decpos--;
								}
								StrCopy(decpos, txt);
								if (txtlen == 2) {
									StrCat(decpos, "0");		// Supply zero for trailing decimal point
								}
							}
						} else if (stoneDDdotD) {
						
							/*	Display units are stones and the user has written
								"dd.d" or "dd.".  Since we never use decimal stones,
								we know such a specification must imply replacement
								of the pounds in the previous entry with the new
								value, further assuming that few infants weighing less
								than one stone will be among our user community.  */
							
							decpos = StrChr(copiedField, ' ');
							
							/*	The conditionals below are purely defensive sanity
								checks.  They should always be true for any value
								in pounds and stones edited into a previous log
								entry.  */
								
							if (decpos != NULL) {
								decpos++;
								if (isDigit(*decpos)) {
									StrCopy(decpos, txt);
									if (txtlen == 3) {
										StrCat(decpos, "0");	// Supply zero for trailing decimal point
									}
								}								
							}
						}
						txt = copiedField;
						break;
					}
				}
				break;
				
			case cRung:
				for (i = RowDay - 1; i >= 0; i--) {
					if (m->rungArray[i] != RUNG_ABSENT) {
						StrPrintF(copiedField, "%d", m->rungArray[i]);
						txt = copiedField;
						break;
					}
				}
				break;
				
			case cComment:
				for (i = RowDay - 1; i >= 0; i--) {
					if (m->commentArray[i] != 0) {
						/*	We have no alternative but to copy the comment we wish to
							duplicate to a temporary buffer on the stack due to the
							fact that adding a comment to the unpacked month may cause
							it to move in memory when reallocating the buffer to include
							the new comment, which would render the absolute pointer to
							the comment in the unpacked log to be invalid.  So don't
							try "optimising" this by setting txt directly to the
							comment we're duplicating.  */
						StrCopy(copiedField, ((char *) m) + m->commentArray[i]);
						txt = copiedField;
						break;
					}
				}
				break;
		}
	}

	switch (col) {
		case cWeight:			// Weight
			{
				Char *decimal;
				weight w = 0;
				Boolean newTrend = false;
				char weightField[24];

				assert(StrLen(txt) < (sizeof weightField));
				StrCopy(weightField, txt);
				txt = weightField;
				if (prefs.displayUnit == wStone) {
					/*	When units are set to stones, accept entries of
						stones and pounds separated by a space.  An
						entry without a space is taken to be a round
						number of stones.  Pounds following stones may
						contain decimal places.  */
					decimal = StrChr(txt, ' ');
					if (decimal != NULL) {
						*decimal = 0;
					}
					NumIn(txt);
					w = StrAToI(txt) * 14 * 100;
					
					//	If a space was present, scan pounds and fraction
					
					if (decimal != NULL) {
						txt = decimal + 1;
						NumIn(txt);
						decimal = StrChr(txt, '.');
						
						if (decimal != NULL) {
							*decimal = 0;
						}
						w += StrAToI(txt) * 100;
						if (decimal != NULL) {
							decimal++;
							while (StrLen(decimal) > 2) {
								decimal[StrLen(decimal) - 1] = 0;
							}
							if (StrLen(decimal) < 2) {
								StrCat(decimal, "0");
							}
							w += StrAToI(decimal);
						}
					}
				} else {
					w = 0;
					
					if (isDigit(txt[0])) {				
						NumIn(txt);
						decimal = StrChr(txt, '.');
						
						if (decimal != NULL) {
							*decimal = 0;
						}
						w = StrAToI(txt) * 100;
						if ((decimal != NULL) && 
							isDigit(decimal[1])) {
							decimal++;
							while (StrLen(decimal) > 2) {
								decimal[StrLen(decimal) - 1] = 0;
							}
							if (StrLen(decimal) < 2) {
								StrCat(decimal, "0");
							}
							w += StrAToI(decimal);
						}
					}
				}
				m->weightArray[RowDay] = cvtWeight(w, prefs.displayUnit, (weightunit) m->weightUnit);
				MarkFieldDirty(RowDay, col);
				if (m->trendCarryForward == 0) {
					newTrend = true;
				} else {
					int n;
					
					for (n = RowDay + 1; n < m->days; n++) {
						if (m->weightArray[n] != 0) {
							newTrend = true;
							break;
						}
					}
				}
				if (newTrend) {
					trendChange = true;
					updateTrend(m);
				}
			}
			break;
			
		case cRung:				// Rung
			{
				UInt16 nr = RUNG_ABSENT;
				
				if ((txt[0] >= '0') && (txt[0] <= '9')) {
					nr = StrAToI(txt);
				}
				if ((nr < 1) || (nr > RUNG_MAX)) {
					nr = RUNG_ABSENT;
				}
				m->rungArray[RowDay] = (rung) nr;
				MarkFieldDirty(RowDay, col);
			}
			break;
			
		case cComment:			// Comment
			{
				struct hdMonth *nm = hdAddComment(m, RowDay, txt);
				
				assert(nm != NULL);
				m = nm;
				MarkFieldDirty(RowDay, col);
			}			
			break;
			
#ifndef NDEBUG
		default:
			assert(false);
			break;
#endif
	}
	return trendChange;
}

//	createField  --  Dynamically cobble up a field for editing an item in the table

static FieldPtr createField(TablePtr tp, UInt16 row, UInt16 col, RectanglePtr extents,
							UInt16 maxchars, Char *contents)
{
	FormPtr frm = FrmGetActiveForm();
	FieldPtr f = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MonthlyLogFloatingField));
	MemHandle hContent;
	Char *pContent;
	FieldAttrType fattr;
	
	FldSetBounds(f, extents);
	
	//	Need to plug some fields to debug ROM doesn't squawk in ValidateField
	
	f->textLen = f->textBlockSize = f->maxChars =
				 f->selFirstPos = f->selLastPos =
				 f->insPtXPos = f->insPtYPos = 0;
	f->text = NULL;
	f->textHandle = 0;
	f->attr.singleLine = f->attr.dynamicSize = f->attr.dirty =
						 f->attr.underlined = f->attr.hasScrollBar = 0;

	FldGetAttributes (f, &fattr);
	fattr.autoShift = (col == cComment) ? 1 : 0;
	fattr.numeric == (col == cComment) ? 0 : 1;
	fattr.insPtVisible = 1;
	fattr.singleLine = 1;
	FldSetAttributes (f, &fattr);
	
	f->maxChars = maxchars;
	
	hContent = MemHandleNew(maxchars + 2);		//	Must be long enough for maximum length
	pContent = MemHandleLock(hContent);
	
	//	Recover pointers since allocation may have moved them
	
	frm = FrmGetActiveForm();
	f = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MonthlyLogFloatingField));

	/*	Watch out!  You might think this test is overkill and could
		be replaced by using StrNCopy() instead of StrCopy() below.
		But bloody StrNCopy doesn't guarantee the destination string
		is zero-terminated!  So we lop the contents argument to the
		maximum field length (which is probably a good idea anyway) if
		it's too long to guarantee we'll never overwrite the field's
		text buffer.  */
		
	if (StrLen(contents) > maxchars) {
		contents[maxchars] = 0;
	}	
	StrCopy(pContent, contents);
	MemPtrUnlock(pContent);
	FldSetTextHandle(f, (MemHandle) hContent);
	f = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MonthlyLogFloatingField));
	WinEraseRectangle(&f->rect, 0);
	FldSetSelection(f, 0, 0);					// Clear selected text
	f->insPtYPos = 0;
	FldSetInsPtPosition(f, 0);					// Set insertion point to start
	FldSetUsable(f, true);
	
	return f;
}

//	fieldEvent  --  Pass on event to active field

static void fieldEvent(EventPtr event)
{
	EventType ef;
	FormPtr f = FrmGetActiveForm();

	assert(currentFld != NULL);
	currentFld = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogFloatingField));	
	EvtCopyEvent(event, &ef);
	ef.eType = fldEnterEvent;
	ef.data.fldEnter.fieldID = currentFld->id;
	ef.data.fldEnter.pField = currentFld;
	
	FldHandleEvent(currentFld, &ef);

#ifdef GNARLY	
	/*	FldHandleEvent may have allocated memory which caused the table
		and/or field pointers to change.  Recover them, then refresh
		the field embedded in the table.  */
	
	{	
		TablePtr tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
		currentFld = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogFloatingField));
		tp->currentField = *currentFld;
	}
#endif
}

//	editField  --  Activate table field for editing when tapped

static void editField(EventPtr event, TablePtr tp, UInt16 row, UInt16 col)
{
	FieldPtr fld;
	FormPtr f = FrmGetActiveForm();
	RectangleType r;
	UInt16 max;
	char ftext[256];
	
	tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
	tp->attr.editing = true;
	tp->currentRow = row;
	tp->currentColumn = col;
	TblGetItemBounds(tp, row, col, &r);
	formatField(ftext, row, col, &max);
	fld = createField(tp, row, col, &r, max, ftext);
	
	//	Recover form and table pointers which may have moved after createField
	f = FrmGetActiveForm();
	tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));

	FldSetSelection(fld, 0, 0);
	f = FrmGetActiveForm();
	tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
	FrmSetFocus(f, FrmGetObjectIndex(f, MonthlyLogFloatingField));
	FldGrabFocus(fld);
	FldDrawField(fld);
	currentFld = fld;
	editRow = row;
	editCol = col;
	editing = true;

	fieldEvent(event);
	max = FldGetTextLength(fld);
	if (max > 0) {
		FldSetSelection(fld, 0, max);
	}
	tp->currentField = *fld;
}

//	quitField  --  Stop editing field

static void quitField(void)
{
	FormPtr f = FrmGetActiveForm();
	TablePtr tp;
	Boolean trendChanged;
	
	if (editing) {
		tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
		currentFld = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogFloatingField));
#ifdef GNARLY
		currentFld->textLen = StrLen(FldGetTextPtr(currentFld));
		//	Beats me how this gets of of sync, but somehow it can (only on 3.3 ROM)
		if (currentFld->insPtXPos > currentFld->textLen) {
			currentFld->insPtXPos = currentFld->textLen;
		}
		tp->currentField = *currentFld;
#endif
		FldSetSelection(currentFld, 0, 0);			// Clear selected text
		FldSetUsable(currentFld, false);			// Disable field input
		FldReleaseFocus(currentFld);				// Release input focus

		if (FldDirty(currentFld)) {
			trendChanged = updateFromField(FldGetTextPtr(currentFld), editRow, editCol);
			dirty = true;							// Set in-memory copy dirty
			/*	updateFromField(), which can allocate additional memory
				for a new comment may have caused a compaction which
				changed the addresses of the form and/or table.  Refresh
				these addresses so they're valid in the following code.  */
			f = FrmGetActiveForm();
			tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
			currentFld = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogFloatingField));
		}
#ifdef GNARLY
		tp->currentField = *currentFld;
#endif
		if ((currentFld->textHandle != 0) && (currentFld->text != NULL)) {
			FldFreeMemory(currentFld);				// Free text editing buffer
		}
#ifdef GNARLY
		tp->currentField = *currentFld;
#endif
		tp->attr.editing = false;					// Mark table no longer editing
		TblUnhighlightSelection(tp);
		if (trendChanged) {
			TblMarkTableInvalid(tp);				// Trend changed--repaint entire table
		} else {
			TblMarkRowInvalid(tp, editRow);			// Row in which we were editing needs repaint...
		}
		TblRedrawTable(tp);							// ...so repaint it, already

		currentFld = NULL;							// Nothing is now being edited
		editRow = -1;
		editCol = -1;
		editing = false;
	}
	clearUnlock();									// Clear temporary row unlock, if any
}

/*	lightCone  --  Determine if a given day is in the past, today, or future:

						-1	Past
						 0	Today
						 1	Future
						 
				   The day argument is 0-30, as used internally to index logs.
*/

static int lightCone(int day)
{
		DateTimeType dt;

		TimSecondsToDateTime(TimGetSeconds(), &dt);
		
		if ((m->year > dt.year) ||
			((m->year == dt.year) && ((m->month + 1) > dt.month)) ||
			((m->year == dt.year) && ((m->month + 1) == dt.month) && ((day + 1) > dt.day))) {
			return 1;					// Future
		}
		
		if ((m->year == dt.year) && ((m->month + 1) == dt.month) &&
			((day + 1) == dt.day)) {
			return 0;					// Today
		}
		
		return -1;						// Past	
}

/*	fieldProtected  --  If field protection is enabled in the preferences,
						refuse to open for editing:
						
							Fields prior to today with nonblank values,
								unless they're "dirty" (have been edited
								in this session)
							Fields with dates in the future
*/

static Boolean fieldProtected(TablePtr tp, UInt16 row, UInt16 col)
{
	if (prefs.protectEntries) {
		char ftxt[256];
		UInt16 fmax;
		int cone = lightCone(RowDay);
		
		if (cone > 0) {
			return true;				// Date is in the future: all fields protected
		}
		if (cone == 0) {
			return false;				// Date is today: all changes permitted
		}
		
		//	Permit changes if the field is temporarily unlocked
		
		if (row == tempUnlock) {
			return false;
		}
		
		/*	Don't allow changes to fields in the past unless the field is
			dirty--it's been modified in this log editing session.  */
			
		if (IsFieldDirty(RowDay, col)) {
			return false;
		}

		if (col == cFlag) {
			/*	Well, actually, since we can't distinguish an unchecked flag from
				one never modified this is a bad idea because it prevents entering
				flags logged on paper.  So, disable it for now until some better
				idea surfaces.  */
#ifdef FLAGPAST
			return true;
#else
			return false;
#endif
		}
		
		//	Date is in the past.  Only allow changes to blank fields
		
		formatField(ftxt, row, col, &fmax);
		if (StrLen(ftxt) > 0) {
			return true;
		}
	}
	return false;
}

//	goToDayInMonth  --  Make sure a given day in the month is visible

static void goToDayInMonth(int dday, Boolean editWeight)
{
	FormPtr f;
	TablePtr tp;
	EventType evt;
	
	/*	Okay, current month is displayed.  If current day
		isn't visible, scroll so it is.  */
		
	quitField();
	f = FrmGetActiveForm();
	if (!tablePainted || (dday < topDay) || (dday >= (topDay + tableRows))) {
		topDay = dday - (tableRows / 2);
		if (topDay < 1) {
			topDay = 1;
		}
		
		/*	If this would scroll off the end of the month, back
			up do it doesn't.  */
			
		if ((topDay + tableRows) >= m->days) {
			topDay = m->days - (tableRows - 1);
		}
		setScrollbar();
		scrollTable();
	}
	
	
	if (editWeight) {
		tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
		
		evt.eType = tblEnterEvent;
		evt.data.tblEnter.row = dday - topDay;
		evt.data.tblEnter.column = cWeight;
		evt.data.tblEnter.pTable = tp;
	
		FrmDispatchEvent(&evt);
	}
}

//	getFocusObject  --  Get pointer to object with focus

static FieldPtr getFocusObject(void)
{
	FormPtr frm;
	UInt16 focus;
	FormObjectKind objType;
	
	frm = FrmGetActiveForm();
	focus = FrmGetFocus(frm);
	if (focus == noFocus) {
		return NULL;
	}
		
	objType = FrmGetObjectType(frm, focus);
	
	if (objType == frmFieldObj) {
		return FrmGetObjectPtr(frm, focus);
	} else if (objType == frmTableObj) {
		currentFld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, MonthlyLogFloatingField));
		return currentFld;
	}
	return NULL;
}

/*	monthLogHandleEvent  --  Handle event while the month log form
							 is displayed.  */
							  

static Boolean monthLogHandleEvent(EventPtr event)
{
	Boolean handled;
#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif
	handled = false;
	
	switch (event->eType) {
	  	
  		case frmOpenEvent:
  			{
	  			FormPtr f = FrmGetActiveForm();

	  			/*	If recordRung is set, plug the requested rung into
	  				today's rung field.  */
	  				
	  			if (recordRung != 0) {
					DateTimeType dt;

					TimSecondsToDateTime(TimGetSeconds(), &dt);
					assert((m->year == dt.year) && ((m->month + 1) == dt.month));
					if (m->rungArray[dt.day - 1] != recordRung) {
						m->rungArray[dt.day - 1] = recordRung;
						MarkFieldDirty(dt.day - 1, cRung);
						dirty = true;
					}
					recordRung = 0;
	  			}
	
				/*	If we're directed to go to today from another form,
					fake a tap on the "Today" button.  */
				
				if (goToToday) {
					showToday = true;
					goToToday = false;
				}
				
				//	If we're showing (but not editing) today, scroll to it
				
				tablePaint = (!showToday) && (showSpecificDay == 0);
				tablePainted = tablePaint;
				FrmDrawForm(f);
				tablePaint = true;
				
				/*	If we've been directed to place a given day in view,
					scroll the table to comply.  */
					
				if (showToday) {
					DateTimeType dt;

					TimSecondsToDateTime(TimGetSeconds(), &dt);
					goToDayInMonth(dt.day, false);
					showToday = false;
				} else if (showSpecificDay != 0) {
					goToDayInMonth(showSpecificDay, false);
					showSpecificDay = 0;
				}
				tablePainted = true;
			}
			handled = true;
			break;
			
		case frmCloseEvent:
			if (m != NULL) {
				quitField();						// Quit any field we're editing
				if (dirty) {
					hdPutMonthUnpacked(m, logCurrentMonth);
					dirty = false;
				}
				hdDisposeMonthUnpacked(m);
				m = NULL;
			}
			MenuEraseStatus(NULL);
			FrmEraseForm(FrmGetActiveForm());
			FrmDeleteForm(FrmGetActiveForm());
			handled = true;
			goto really_done;
			
		case frmSaveEvent:							// Save dirty month in database prior to find
			if (m != NULL) {
				quitField();						// Quit any field we're editing
				if (dirty) {
					hdPutMonthUnpacked(m, logCurrentMonth);
					dirty = false;
				}
			}
			handled = true;
			break;
			
		//	Previous and next month repeating controls (in title bar)
			
		case ctlRepeatEvent:
			switch(event->data.ctlRepeat.controlID) {
				case MonthlyLogPreviousMonthRepeating:	// Previous month
					logGotoMonth = logCurrentMonth - 1;
changeMonth:		FrmGotoForm(MonthlyLogForm);
					handled = true;
					break;
					
				case MonthlyLogNextMonthRepeating:	// Next month
					logGotoMonth = logCurrentMonth + 1;
					goto changeMonth;
			}
			break;
			
		case ctlSelectEvent:
			switch(event->data.ctlSelect.controlID) {
				case MonthlyLogDoneButton:			// Done button: Back to year view of current year
					if (m == NULL) {
						monthLogShowToday();
					} else {
						yearGotoYear = m->year;
						FrmGotoForm(YearViewForm);
					}
					handled = true;
					break;
										
				case MonthlyLogTodayButton:			// Today button: Show entry for today
					{
						DateTimeType dt;
	
						TimSecondsToDateTime(TimGetSeconds(), &dt);
						if ((m->year != dt.year) || (m->month != (dt.month - 1))) {
							goToToday = true;
							hdEnsureMonthPresent(dt.year, dt.month - 1);
							logGotoMonth = hdFindMonth(dt.year, dt.month - 1);
							goto changeMonth;
						}
						goToDayInMonth(dt.day, m->weightArray[dt.day - 1] == 0);
					}
					handled = true;
					break;
					
				case MonthlyLogChartButton:			// Chart button: Show chart for month
					logGotoMonth = logCurrentMonth;
					FrmGotoForm(MonthlyChartForm);
					handled = true;
					break;
					
				case MonthlyLogTrendButton:			// Trend button: Show trend analysis
					FrmGotoForm(TrendAnalysisForm);
					handled = true;
					break;
					
				case MonthlyLogRungButton:			// Rung stealth button: Show exercise ladder
					FrmGotoForm(ExerciseForm);
					handled = true;
					break;					
			}
			break;
			
		case tblEnterEvent:						// Handle entry into editable fields
			{
				UInt16 row = event->data.tblEnter.row,
					 col = event->data.tblEnter.column;
				TablePtr tp = event->data.tblEnter.pTable;
				
				/*	If a row is temporarily unlocked, re-lock it when the user taps
					in a different row.  */
					
				if ((tempUnlock >= 0) && (tempUnlock != row)) {
					clearUnlock();
				}
				if ((tempUnlockFirst >= 0) && ((tempUnlockFirst != row) || (col != cDate))) {
					tempUnlockFirst = -1;
				}
				
				/*	If we're editing and the user taps in a different cell, stop
					editing the previous cell.  */
				
				if (editing && (editRow != row || editCol != col)) {
					quitField();
				}
				
				if (columnEditable(col)) {
					if (editing) {
						fieldEvent(event);		// Forward event to active field
					} else {
						if (!fieldProtected(tp, row, col)) {
							editField(event, tp, row, col);
						}
					}
					handled = true;
				} else if (col == cFlag) {
					handled = fieldProtected(tp, row, col);
					goto all_done;
				} else if ((col == cDate) && prefs.protectEntries) {
				
					/*	The following logic permits the user to temporarily remove
						protection from a row in a table prior to the current date
						by tapping twice (without any other intervening actions)
						in its Date cell.  The Date is redrawn inverted, and the user
						may then modify a single field in that row, presumably to
						correct a typo.  After that field is closed or the user
						taps any other row, the temporary unlocking is removed and
						the date is redrawn non-inverted.  If you need to modify more
						than one item on a past date, you need to unlock the row
						before editing each item (how likely is it you'll have more
						than one typo in a day's log item?).  */
				
					if ((tempUnlockFirst >= 0) && (tempUnlockFirst == row)) {
						RectangleType r;
						
						if (lightCone(RowDay) >= 0) {
							handled = true;				// Can't temporarily unlock today or in the future
							goto all_done;
						}
						tempUnlock = row;				// Mark cell temporarily unlocked
						TblGetItemBounds(tp, row, col, &r);
						drawDate(tp, row, col, &r);		// Redraw date to mark row unlocked
					} else {
						tempUnlockFirst = row;
					}
				} else {
					handled = true;
					goto all_done;
				}
			}
			break;
			
		case tblSelectEvent:					// Handle tblSelectEvent from checkboxes in table
			{
				UInt16 row = event->data.tblExit.row,
					 col = event->data.tblExit.column;
				 
				if (col == cFlag) {
					m->flagsArray[row + (topDay - 1)] = !m->flagsArray[row + (topDay - 1)];
					dirty = true;
				}
				handled = true;
			}
			break;

#ifdef SCROLLTRACK
		/*	This code handles on the fly scrolling of the table as the
			scroll bar is moved.  This is disabled by default because
			as of the Palm IIIx era the CPU isn't fast enough to repaint
			the whole table smoothly as you scroll.  */			
		case sclRepeatEvent:
			{
				quitField();
				topDay = event->data.sclRepeat.newValue + 1;
				scrollTable();
				handled = true;
			}
			break;
#else			
		case sclExitEvent:
			{
				quitField();
				topDay = event->data.sclExit.newValue + 1;
				scrollTable();
				handled = true;
			}
			break;
#endif			
		
		case keyDownEvent:
			if (editing && (currentFld != NULL)) {
				UInt16 c = event->data.keyDown.chr;
				
				if ((c == backspaceChr) || (c == leftArrowChr) ||
					(c == rightArrowChr) ||
					(isAscii(c) && isPrint(c))) {
					
					/*	If we're editing the rung or weight fields, accept
						only decimal digits and a decimal point.  We permit a
						decimal point in the rung field to denote copying the
						last non-blank rung in the month.  The locale's decimal
						point character may be entered in the weight field only,
						and is never used to copy a previous field--that always
						requires a period.  */
					   
					if (isGraph(c) &&
						(((editCol == cWeight) || (editCol == cRung)) && (!(isDigit(c) ||
							(c == '.') || ((editCol == cWeight) && (c == numberDecimal)))))) {
						/*	Must set handled so default field event
							handler won't process character anyway.  */
						handled = true;
						break;
					}
					
					/*	If we're editing the comment field and we've
						just entered an initial letter, capitalise
						it.  Maybe someday we'll learn why autoShift
						won't do this for us (though it seems to do
						it for the built-in applications). */
					if ((editCol == cComment) &&
						(FldGetTextLength(currentFld) == 0) &&
						isLower(c)) {
						event->data.keyDown.chr = toUpper(c);
					}
#ifdef GNARLY
					//	Regenerate text length in field in case PalmOS gets it wrong
					currentFld->textLen = StrLen(FldGetTextPtr(currentFld));
#endif
					handled = FldHandleEvent(currentFld, event);
				}
#ifdef GNARLY
				/*	Even if we didn't handle the character, if it's an ASCII
					character (' ' <= c <= 255) we want to say we handled it
					lest PalmOS get the idea *it* should do something about
					it.  */
				if (isAscii(c) && (c >= ' ')) {
					handled = true;
				}
#endif
				if (handled) {
					break;
				}
			
			/*	One more gimmick to expedite data entry.  If no field in the table
			    is selected and the user writes a digit or period in the Graffiti area while
			    today's weight field is blank, open today's weight and enter the
			    digit as the first character.  We don't have to enter the digit
			    manually in the field--simply leaving the event as "un-handled"
			    automagically passes it to the field event handler somehow.  Of
			    course this applies only if the log displayed is that which
			    contains today's date.  */
				
			} else if (isDigit(event->data.keyDown.chr) ||
					   (event->data.keyDown.chr == '.')) {
				DateTimeType dt;

				TimSecondsToDateTime(TimGetSeconds(), &dt);
				if ((m->year == dt.year) && (m->month == (dt.month - 1)) &&
					m->weightArray[dt.day - 1] == 0) {
					goToDayInMonth(dt.day, true);
				}
			}
		
			switch (event->data.keyDown.chr) {
				case linefeedChr:						// Line feed: exit field editing
					quitField();
					handled = true;
					break;
					
				case pageUpChr:
					{	int n = topDay;
					
						quitField();					// Close any field we're editing
						n -= tableRows;
						if (n < 1) {
							n = 1;
						}
						topDay = (UInt16) n;
						setScrollbar();
						scrollTable();
						handled = true;
					}
					break;
					
				case pageDownChr:
					quitField();						// Close any field we're editing
					topDay += tableRows;
					if (topDay > ((m->days - tableRows) + 1)) {
						topDay = (m->days - tableRows) + 1;
					}
					setScrollbar();
					scrollTable();
					handled = true;
					break;
			}
			break;
						
		case menuEvent:
			{
				FieldPtr f = getFocusObject();

/*	Constructor (the MetroWerks Palm resource editor) has developed the
	absolutely infuriating tactic of changing the #defines for the menu
	items for the monthly log to different symbols every time you edit
	one of the menus.  I suspect this is due to its general confusion
	over the relationship between menu bars and menus (allowing you to
	edit menus within the menu bar inspector).  Anyway, psychoanalysis
	having failed, let's pop a Prozac into the program and, if the correct
	symbols aren't defined, find out which ones *are* and define the
	cases in the menu code to whichever Constructor spat out the last time
	we edited the menu.  Why "Gazoom"?  Because Constructor got befuddled
	about name uniqueness and generated definitions of different objects
	with the same name, so I had to rename the menu to avoid the
	ambiguity.  */
					
#ifndef LogEditUndo
#ifdef LogUndo
#define LogEditUndo LogUndo
#define LogEditCut LogCut
#define LogEditCopy LogCopy
#define LogEditPaste LogPaste
#define LogEditPreferences LogPreferences
#define LogEditSelectAll LogSelectAll
#define LogEditKeyboard LogKeyboard
#define LogEditGraffitiHelp LogGraffitiHelp
#endif

#ifdef GazoomUndo
#define LogEditUndo GazoomUndo
#define LogEditCut GazoomCut
#define LogEditCopy GazoomCopy
#define LogEditPaste GazoomPaste
#define LogEditPreferences GazoomPreferences
#define LogEditSelectAll GazoomSelectAll
#define LogEditKeyboard GazoomKeyboard
#define LogEditGraffitiHelp GazoomGraffitiHelp
#endif

#ifdef ViewInLogChart
#define ViewChart ViewInLogChart
#define ViewTrendAnalysis ViewInLogTrendAnalysis
#define ViewThisYear ViewInLogThisYear
#define ViewAllYears ViewInLogAllYears
#define ViewOtherMonths ViewInLogAllMonths
#define ViewExerciseLadder ViewInLogExerciseLadder
#define ViewDietCalculator ViewInLogDietCalculator
#define ViewKilograms ViewInLogKilograms
#define ViewPounds ViewInLogPounds
#define ViewStones ViewInLogStones
#define ViewAboutEatWatch ViewInLogAboutEatWatch
#endif
#endif
				
				switch (event->data.menu.itemID) {
					case LogEditUndo:
						if (f != NULL) {
							FldUndo(f);
						}
						handled = true;
						break;
						
					case LogEditCut:
						if (f != NULL) {
							FldCut(f);
						}
						handled = true;
						break;
						
					case LogEditCopy:
						if (f != NULL) {
							FldCopy(f);
						}
						handled = true;
						break;
						
					case LogEditPaste:
						if (f != NULL) {
							FldPaste(f);
						}
						handled = true;
						break;
						
					case LogEditSelectAll:
						if (f != NULL) {
							FldSetSelection(f, 0, FldGetTextLength(f));
						}
						handled = true;
						break;
						
					case LogEditPreferences:
						FrmGotoForm(PreferencesForm);
						handled = true;
						break;
						
					case LogEditKeyboard:
						SysKeyboardDialog(kbdDefault);
						handled = true;
						break;
						
					case LogEditGraffitiHelp:
						SysGraffitiReferenceDialog(referenceDefault);
						handled = true;
						break;
					
					case MonthlyLogChartingThisMonth:	
					case ViewChart:
						logGotoMonth = logCurrentMonth;
						FrmGotoForm(MonthlyChartForm);
						handled = true;
						break;
						
					case ViewTrendAnalysis:
						FrmGotoForm(TrendAnalysisForm);
						handled = true;
						break;
						
					case ViewThisYear:
						if (m != NULL) {
							yearGotoYear = m->year;
							FrmGotoForm(YearViewForm);
						}
						handled = true;
						break;						
					
					case ViewAllYears:
						FrmGotoForm(YearListForm);
						handled = true;
						break;
						
					case ViewOtherMonths:
						FrmGotoForm(MonthsForm);
						handled = true;
						break;
						
					case ViewExerciseLadder:
						FrmGotoForm(ExerciseForm);
						handled = true;
						break;
						
					case ViewDietCalculator:
						FrmGotoForm(DietCalcForm);
						handled = true;
						break;
						
					case ViewKilograms:
						{
							FormPtr f;
							TablePtr tp;

							quitField();
							f = FrmGetActiveForm();
							tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
							prefs.displayUnit = wKilogram;
							TblEraseTable(tp);
							TblDrawTable(tp);
						}
						handled = true;
						break;
						
					case ViewPounds:
						{
							FormPtr f;
							TablePtr tp;

							quitField();
							f = FrmGetActiveForm();
							tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
							prefs.displayUnit = wPound;
							TblEraseTable(tp);
							TblDrawTable(tp);
						}
						handled = true;
						break;
						
					case ViewStones:
						{
							FormPtr f;
							TablePtr tp;

							quitField();
							f = FrmGetActiveForm();
							tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
							prefs.displayUnit = wStone;
							TblEraseTable(tp);
							TblDrawTable(tp);
						}
						handled = true;
						break;
				
					case ViewAboutEatWatch:
						aboutDialogue();
						break;
						
					case MonthlyLogChartingLast30Days:
						historyMonths = 1;
histChart:;				{
							DateTimeType dt;
							
							dt.year = m->year;
							dt.month = m->month + 1;
							dt.day = m->days;
							dt.hour = dt.minute = dt.second = 0;
							historyGoTo = TimDateTimeToSeconds(&dt);
						}
						FrmGotoForm(HistoricalChartForm);
						handled = true;
						break;
						
					case MonthlyLogChartingQuarter:
						historyMonths = 3;
						goto histChart;
						
					case MonthlyLogChartingSixMonths:
						historyMonths = 6;
						goto histChart;
						
					case MonthlyLogChartingYear:
						historyMonths = 12;
						goto histChart;

					case MonthlyLogChartingEntireHistory:
						historyMonths = 0;
						goto histChart;
				}
			}
	}

all_done:;

#ifdef GNARLY

	/*	If we're editing a field, first make sure the text length and
		insertion point are valid and fix them if they point beyond the
		end of the string, then copy the field into the FieldType
		structure in the table so that zany code like Find can grab
		it from there.  This is required to avoid squawks from the
		PalmOS 3.3 debug ROM.  */

	if (currentFld != NULL) {
		FormPtr f = FrmGetActiveForm();
		TablePtr tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
		Char *ftxt;
		
		if (currentFld->text == NULL) {
			currentFld->textLen = 0;
		} else {
			currentFld->textLen = StrLen(currentFld->text);
		}
		ftxt = FldGetTextPtr(currentFld);
		
		if (ftxt != NULL) {
			currentFld->textLen = StrLen(ftxt);
		} else {
			currentFld->textLen = 0;
		}
		if (currentFld->insPtXPos > currentFld->textLen) {
			currentFld->insPtXPos = currentFld->textLen;
		}
		tp->currentField = *currentFld;
	}
#endif

really_done:;

#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
	return handled;
}

//	drawDate  --  Draw date column items

static void drawDate(MemPtr table, Int16 row, Int16 col, RectanglePtr rect)
{
	int mday, wday;
	char wd[6], ftxt[20];
#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif

	if (tablePaint) {
		if (RowDay < m->days) {
			FontID fid;
			Boolean today;
			
			mday = topDay + row;
			wday = (m->firstDay + RowDay) % 7;
			today = lightCone(RowDay) == 0;
			stringListItem(DayAbbreviationsStringList, wday, wd, sizeof wd);
			wd[2] = 0;
			StrPrintF(ftxt, "%d %s", mday, wd);
			if (today) {
				fid = FntSetFont(boldFont);
			}
			if (row == tempUnlock) {
				drawInFieldInverted(ftxt, rect);
			} else {
				drawInField(ftxt, rect);
			}
			if (today) {
				FntSetFont(fid);
			}
		}
	}
#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
}

//	drawWeight  --  Draw weight column items

static void drawWeight(MemPtr table, Int16 row, Int16 col, RectanglePtr rect)
{
	char txt[12];
#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif

	if (tablePaint) {
		if ((RowDay < m->days) && (m->weightArray[RowDay] > 0)) {
			editWeight(txt, cvtWeight(m->weightArray[RowDay], (weightunit) m->weightUnit, prefs.displayUnit));
		} else {
			StrCopy(txt, "___");
		}
		drawInFieldRight(txt, rect);
	}
	
#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
}

//	drawVar  --  Draw variance column items

static void drawVar(MemPtr table, Int16 row, Int16 col, RectanglePtr rect)
{
	char txt[12];
	char *sgn;
	signed long var, whole, frac;
#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif

	if (tablePaint) {
		if ((RowDay < m->days) && (m->weightArray[RowDay] > 0)) {
			var = ((signed long) WeightToTrend(m->weightArray[RowDay])) -
				  ((signed long) m->trendArray[RowDay]);
			var = cvtLong(var, (weightunit) m->weightUnit, prefs.displayUnit);
			sgn = (var < 0) ? "-" : "+";
			if (var < 0) {
				var = -var;
			}
			var += 50000L;
			whole = var  / 1000000L;
			frac = (var / 100000L) % 10;
			if (whole == 0 && frac == 0) {
				sgn = "";
			}
			StrPrintF(txt, "%s%ld.%ld", sgn, whole, frac);
			NumOut(txt);
			if (useColour) {
				WinPushDrawState();
				if (sgn[0] != 0) {
					WinSetTextColor((sgn[0] == '+') ? PrefColourIndex(logPlusVarianceColour) :
													  PrefColourIndex(logMinusVarianceColour));
				}
			}
			drawInFieldRight(txt, rect);
			if (useColour) {
				WinPopDrawState();
			}
		}
	}
	
#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
}

//	drawRung  --  Draw rung column items

static void drawRung(MemPtr table, Int16 row, Int16 col, RectanglePtr rect)
{
	char txt[12];
#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif

	if (tablePaint) {
		if ((RowDay < m->days) && (m->rungArray[RowDay] != RUNG_ABSENT)) {
			StrPrintF(txt, "%d", m->rungArray[RowDay]);
			drawInFieldRight(txt, rect);
		}
	}
	
#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
}

//	drawComment  --  Draw comment column items

static void drawComment(MemPtr table, Int16 row, Int16 col, RectanglePtr rect)
{
#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif

	if (tablePaint) {
		if ((RowDay < m->days) && (m->commentArray[RowDay] != NULL)) {
			drawInField(((char *) m) + m->commentArray[RowDay], rect);
		}
	}
	
#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
}

/*	monthLogShowToday  --  Load (creating if necessary) the monthly log form
						   for the present month and scroll so today's entry
						   is visible.  */
						   
void monthLogShowToday(void)
{
	DateTimeType dt;
	
	TimSecondsToDateTime(TimGetSeconds(), &dt);
	hdEnsureMonthPresent(dt.year, dt.month - 1);
	
	logGotoMonth = hdFindMonth(dt.year, dt.month - 1);
	showToday = true;
	FrmGotoForm(MonthlyLogForm);
}

/*	monthLogLoad  --  Initialise month log form and
					  display.  Called when ApplicationHandleEvent
					  receives a frmLoadEvent for the month log
					  form. */

void monthLogLoad(FormPtr f)
{
	TablePtr tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
	char monthName[MAX_MONTH_NAME];
	UInt16 i, j, nr;
	
	logUnlocked = true;
	if (hidePrivateData && prefs.privateLogs) {
		FrmAlert(PrivateLogAlert);
		FrmGotoForm(PreferencesForm);
		logUnlocked = false;
	}
	tempUnlock = tempUnlockFirst = -1;		

	topDay = 1;					// Show start of month
	
	logCurrentMonth = logGotoMonth;
	m = hdGetMonthUnpacked(logCurrentMonth);
	
	//	hdGetMonthUnpacked may have moved form or table
	
	f = FrmGetActiveForm();
	tp = FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogMonthTable));
	
	dirty = false;
	MemSet(fldDirty, sizeof fldDirty, 0);	// Clear field dirty flags
	
	editing = false;
	currentFld = NULL;
	
	//	Hide next and/or previous repeating buttons if we're at the end of the database

	if (logCurrentMonth == 0) {
		CtlHideControl(FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogPreviousMonthRepeating)));
	}
	if (logCurrentMonth == (hdNumMonths() - 1)) {
		CtlHideControl(FrmGetObjectPtr(f, FrmGetObjectIndex(f, MonthlyLogNextMonthRepeating)));
	}
	
	if (m != NULL) {
		FrmSetEventHandler(f, monthLogHandleEvent);
		tableRows = nr = TblGetNumberOfRows(tp);
		
		TblSetCustomDrawProcedure(tp, cDate, drawDate);
		TblSetCustomDrawProcedure(tp, cWeight, drawWeight);
		TblSetCustomDrawProcedure(tp, cVar, drawVar);
		TblSetCustomDrawProcedure(tp, cRung, drawRung);
		TblSetCustomDrawProcedure(tp, cComment, drawComment);
		
		for (i = 0; i < nr; i++) {
			for (j = 0; j < nColumns; j++) {
				if (j == cFlag) {
					UInt16 d = i + (topDay - 1);
					
					TblSetItemStyle(tp, i, j, checkboxTableItem);
					TblSetItemInt(tp, i, j, d < m->days ? m->flagsArray[d] : 0);
				} else {
					TblSetItemStyle(tp, i, j, customTableItem);
				}
			}
		}
	
		for (i = 0; i < nColumns; i++) {
			TblSetColumnUsable(tp, i, true);
			TblSetColumnEditIndicator(tp, i, columnEditable(i) || (i == cFlag));
		}
	
		stringListItem(MonthNamesStringList, m->month, monthName, sizeof monthName);
		StrPrintF(formTitle, "%s %d", monthName, m->year);
#ifdef QUICKTREND
		StrCat(formTitle + StrLen(formTitle), " **QT**");
#endif
		FrmSetTitle(f, formTitle);
		setScrollbar();
	}
}